#include <stdio.h>
#include <windows.h>
#include "privilege.h"

void print_help() {
    printf("Usage:\n"
           "  xln <target file> <hard link name>\n"
           "Examples:\n"
           "  xln C:\\MSDN\\MSDN120.NT4.COL  C:\\MSDN\\MSDN120.COL\n"
          );
    exit(0);
}

BOOLEAN
make_hard_link(
    PWCHAR TargetFN,
    PWCHAR LinkFN
    )
{
    HANDLE          hFile;
    WIN32_STREAM_ID streamId;
    ULONG           bytes_written;
    ULONG           RC;
    LPVOID          lpContext = NULL;
    WCHAR           LinkFullPath[MAX_PATH*2];
    PWCHAR          ptrFN = NULL;
    ULONG           len;

    hFile = CreateFileW(TargetFN,0x100,    // Special privileges
                     FILE_SHARE_READ | FILE_SHARE_WRITE,
                     NULL,
                     OPEN_EXISTING,
                     0,
                     NULL);
    if(hFile == ((HANDLE)-1)) {
        RC = GetLastError();
        printf("Error: cant open target file: %ws (error %#x)\n", TargetFN, RC);
        return FALSE;
    }

    if(!(len = GetFullPathNameW(LinkFN, MAX_PATH*2-1, LinkFullPath, &ptrFN))) {
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error: cant generate full path for link (error %#x)\n", RC);
        return FALSE;
    }

    LinkFullPath[len] = 0;

    streamId.dwStreamId = BACKUP_LINK;
    streamId.dwStreamAttributes = 0;
    streamId.dwStreamNameSize = 0;
    streamId.Size.HighPart = 0;
    streamId.Size.LowPart = (wcslen(LinkFullPath)+1)*sizeof(WCHAR);

    if (!BackupWrite(hFile, (PUCHAR)&streamId, sizeof(WIN32_STREAM_ID)-4, &bytes_written, FALSE, FALSE, &lpContext)) {
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error: cant't create link file (error %#x)\n", RC);
        return FALSE;
    }
    if (!BackupWrite(hFile, (PUCHAR)&LinkFullPath[0], streamId.Size.LowPart, &bytes_written, FALSE, FALSE, &lpContext)){
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error: cant't write link info (error %#x)\n", RC);
        exit(1);
    }
    if (!BackupWrite(hFile, NULL, 0, &bytes_written, TRUE, FALSE, &lpContext)){
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error: cant't commit link info (error %#x)\n", RC);
        return FALSE;
    }

    CloseHandle(hFile);
    return TRUE;
}

int
main (void)
{
    int argc;
    WCHAR** argv;
    WCHAR*  CmdLine;

    printf("NT File System link management utility v1.0 (c) by Alter, 2004\n"
           "Home site http://www.alter.org.ua\n");

    CmdLine = GetCommandLineW();
    argv = CommandLineToArgvW(CmdLine, &argc);

    if(argc < 3) {
        print_help();
    }

    if (!Privilege(SE_BACKUP_NAME, TRUE)) {
        printf("Error: insufficient privileges\n");
        return 1;
    }
    if (!Privilege(SE_RESTORE_NAME, TRUE)) {
        printf("Error: insufficient privileges\n");
        return 1;
    }

    if(!make_hard_link(argv[1], argv[2])) {
        return 1;
    }

    return 0;
}
